import portpicker
from fastapi.responses import Response
import requests
from tensorboard import notebook

from colab_easy_ui.const import is_running_on_colab
import os


class ColabInternalFetcher:
    tb_port = 0

    def __init__(self, project_dir: str, ipython=None, logfile=None) -> None:
        self.project_dir = project_dir
        self.ipython = ipython
        self.logfile = logfile

    def start_tensorboard_colab(self, tb_logs_dir):
        self.tb_port = portpicker.pick_unused_port()
        if self.logfile is not None:
            self.ipython.system_raw(f"tensorboard --port {self.tb_port} --logdir {tb_logs_dir}  --samples_per_plugin images=100,audio=100 >{self.logfile} 2>&1 &")
        else:
            try:
                self.ipython.system_raw(f"tensorboard --port {self.tb_port} --logdir {tb_logs_dir}  --samples_per_plugin images=100,audio=100  &")
            except Exception as e:
                print(e)
        return self.tb_port

    def start_tensorboard_normal(self, tb_logs_dir):
        self.tb_port = portpicker.pick_unused_port()
        notebook.start(f"--port {self.tb_port} --logdir {tb_logs_dir} --samples_per_plugin images=100,audio=100")
        self.tb_port = self.tb_port
        return self.tb_port

    def start_tensorboard(self, project_name: str):
        tb_logs_dir = os.path.join(self.project_dir, project_name, "logs")
        os.makedirs(tb_logs_dir, exist_ok=True)

        if is_running_on_colab() is True:
            tb_port = self.start_tensorboard_colab(tb_logs_dir)
        else:
            tb_port = self.start_tensorboard_normal(tb_logs_dir)
        print("tensorboard port:", tb_port, "logs dir:", tb_logs_dir)

    def _fetch(self, url):
        try:
            response = requests.get(url)
            content_type = response.headers.get("Content-Type")
            return Response(content=response.content, media_type=content_type)
        except Exception as e:
            print("internal fetcher exception:", e)
            return {}

    def _post(self, url, data):
        try:
            response = requests.post(url, data=data)
            content_type = response.headers.get("Content-Type")
            return Response(content=response.content, media_type=content_type)
        except Exception as e:
            print("internal fetcher exception:", e)
            return {}

    def get_runs(self):
        if self.tb_port == 0:
            return {}
        url = f"http://localhost:{self.tb_port}/data/runs"
        return self._fetch(url)

    def get_scalars_tags(self):
        if self.tb_port == 0:
            return {}
        url = f"http://localhost:{self.tb_port}/data/plugin/scalars/tags"
        return self._fetch(url)

    def get_scalars_scalars(self, run: str, tag: str):
        if self.tb_port == 0:
            return []
        url = f"http://localhost:{self.tb_port}/data/plugin/scalars/scalars?run={run}&tag={tag}"
        return self._fetch(url)

    def get_scalars_scalars_multirun(self, run: str, tag: str):
        if self.tb_port == 0:
            return []
        url = f"http://localhost:{self.tb_port}/data/plugin/scalars/scalars_multirun"

        data = {
            "tag": tag,
            "runs": [run],
        }
        return self._post(url, data)

    def get_images_tags(self):
        if self.tb_port == 0:
            return {}
        url = f"http://localhost:{self.tb_port}/data/plugin/images/tags"
        return self._fetch(url)

    def get_images_images(self, run: str, tag: str, sample: int):
        if self.tb_port == 0:
            return []
        url = f"http://localhost:{self.tb_port}/data/plugin/images/images?run={run}&tag={tag}&sample={sample}"
        return self._fetch(url)

    def get_images_individualImage(self, blob_key: str):
        if self.tb_port == 0:
            return {}
        url = f"http://localhost:{self.tb_port}/data/plugin/images/individualImage?blob_key={blob_key}"
        return self._fetch(url)

    def get_audio_tags(self):
        if self.tb_port == 0:
            return {}
        url = f"http://localhost:{self.tb_port}/data/plugin/audio/tags"
        return self._fetch(url)

    def get_audio_audio(self, run: str, tag: str, sample: int):
        if self.tb_port == 0:
            return []
        url = f"http://localhost:{self.tb_port}/data/plugin/audio/audio?run={run}&tag={tag}&sample={sample}"
        return self._fetch(url)

    def get_audio_individualAudio(self, blob_key: str, content_type: str):
        if self.tb_port == 0:
            return {}
        url = f"http://localhost:{self.tb_port}/data/plugin/audio/individualAudio?blob_key={blob_key}&content_type={content_type}"
        return self._fetch(url)



import fire
import time
import functools
import os
from colab_easy_ui.ColabEasyUI import ColabEasyUI
from colab_easy_ui.plugins.unzip_function import unzip
from colab_easy_ui.plugins.json_save_function import json_save
from colab_easy_ui.Functions import JsonApiFunc

def run_server(
    ipython=None,
    logfile=None,
    port: int | None = None,
    allow_origins: list[str] = ["*"],
    # display=None,
    # download_func=None,
):
    c = ColabEasyUI.get_instance(port=port, allow_origins=allow_origins)
    port = c.port
    print(f"port: {port}")

    # ファイルアップローダ
    c.enable_file_uploader("upload", {"Voice(zip)": "voice.zip"})
    c.register_functions(
        [
            JsonApiFunc(
                id="unzip",
                type="progress",
                display_name="unzip",
                method="GET",
                path="/functions_unzip",
                func=functools.partial(unzip, port=port, zip_path="upload/voice.zip", extract_to="raw_data"),
            ),
            JsonApiFunc(
                id="json_save",
                type="progress",
                display_name="json_save",
                method="GET",
                path="/functions_json_save",
                func=functools.partial(json_save, port=port, output_path="./config.json"),
            ),
        ])
    
    current_file_path = os.path.abspath(__file__)
    current_dir = os.path.dirname(current_file_path)
    static_files_path = os.path.join(current_dir, "frontend/dist")
    c.mount_static_folder("/front", static_files_path)

    port = c.start()
    print(f"open http://localhost:{port}/front")
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("サーバーを終了します")
        c.stop()
        print("サーバーを終了しました")

def wrapped_run_server(
    # port: int,
    ipython=None,
    logfile=None,
    port: int | None = None,
    allow_origins: list[str] = ["*"],
):
    """
    run_serverの戻り値であるportを標準出力に出さないようにするためのラッパ。
    """
    run_server(ipython, logfile, port=port, allow_origins=allow_origins)
    

def main():
    fire.Fire(wrapped_run_server)

if __name__ == "__main__":
    main()



import logging
from pathlib import Path
import warnings
from colab_easy_ui.const import LOG_FILE
warnings.filterwarnings("ignore", message="Duplicate Operation ID")


class CustomFormatter(logging.Formatter):
    def format(self, record):

        # record.module = f"{record.name}:{record.module}".ljust(30)
        record.name = record.name.ljust(10)[:10]
        record.module = record.module.ljust(20)[:20]
        return super().format(record)


class LogSuppressFilter(logging.Filter):
    def filter(self, record):
        return False


class IgnoreConvertRequestFilter(logging.Filter):
    def filter(self, record):
        logfile = f"/{LOG_FILE}" not in record.getMessage()
        return all([logfile])


def setup_logger(logger_name: str, filename: Path, level=logging.DEBUG):
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    # custom_datefmt = "%Y-%m-%d %H:%M:%S"
    custom_datefmt = "%H:%M:%S"
    if level == logging.DEBUG:
        # formatter = CustomFormatter("%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s - %(pathname)s - %(lineno)s", datefmt=custom_datefmt)
        formatter = CustomFormatter("%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s - %(pathname)s - %(lineno)s")
    else:
        formatter = CustomFormatter("%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s - %(module)s - %(lineno)s", datefmt=custom_datefmt)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    file_handler = logging.FileHandler(filename, mode="a", encoding="utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    for logger_name in ["uvicorn", "uvicorn.error", "uvicorn.access"]:
        logger = logging.getLogger(logger_name)
        logger.propagate = False
        logger.addHandler(file_handler)

    logging.getLogger("uvicorn.access").addFilter(IgnoreConvertRequestFilter())



import os
import shutil

from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi import UploadFile, File, Form
from fastapi import APIRouter

from colab_easy_ui.data.Response import EasyFileUploaderResponse


class EasyFileUploader:
    def __init__(self, upload_dir: str):
        self.upload_dir = upload_dir
        os.makedirs(self.upload_dir, exist_ok=True)
        self.allowed_filenames: dict[str, str] | None = None
        self.file_titles: list[str] | None = None
        self.router = APIRouter()
        self.router.add_api_route("/uploader_info", self.get_info, methods=["GET"])
        self.router.add_api_route("/uploader_upload_file", self.post_upload_file, methods=["POST"])
        self.router.add_api_route("/uploader_concat_uploaded_files", self.post_concat_uploaded_file, methods=["POST"])

    def set_allowed_filenames(self, filenames: dict[str, str] | None):  # filenamestitle, filename
        """
        upload 可能なファイルを指定する。

        Args:
            filenames (dict[str, str] | None): "key:ファイルのタイトル"と"val:アップロード可能なファイル名"のペアの辞書。タイトルはアプリケーション上に表示される名前を想定。valが実際に許容されるファイル名。


        """
        self.allowed_filenames = filenames

    def get_info(self):
        try:
            response = EasyFileUploaderResponse(
                status="OK",
                message="",
                allowed_filenames=self.allowed_filenames,
            )
            json_compatible_item_data = jsonable_encoder(response)
            return JSONResponse(content=json_compatible_item_data)
        except Exception as e:
            response = EasyFileUploaderResponse(
                status="NG",
                message=f"Exception: {e}",
            )
            json_compatible_item_data = jsonable_encoder(response)
            return JSONResponse(content=json_compatible_item_data)

    def _sanitize_filename(self, filename: str) -> str:
        if self.allowed_filenames is not None:
            if filename not in self.allowed_filenames.values():
                print(f"filename {filename} is not allowed. Allowed filenames are {self.allowed_filenames}")
                raise RuntimeError
            # else:
            #     print(f"filename {filename} is allowed. Allowed filenames are {self.allowed_filenames}")

        safe_filename = os.path.basename(filename)
        max_length = 255
        if len(safe_filename) > max_length:
            file_root, file_ext = os.path.splitext(safe_filename)
            safe_filename = file_root[: max_length - len(file_ext)] + file_ext

        return safe_filename

    #######################################
    # Upload File
    # TEST:
    #  curl -X POST -F "filename=data1" -F "index=0" -F "file=@test_data/data0.txt" http://localhost:8001/upload_file
    #  curl -X POST -F "filename=data1" -F "index=1" -F "file=@test_data/data0.txt" http://localhost:8001/upload_file
    #  curl -X POST -F "filename=data1" -F "index=2" -F "file=@test_data/data2.txt" http://localhost:8001/upload_file
    #######################################
    def post_upload_file(
        self,
        file: UploadFile = File(...),
        filename: str = Form(...),
        index: int = Form(...),
    ):
        try:
            res = self._upload_file(self.upload_dir, file, filename, index)
            json_compatible_item_data = jsonable_encoder(res)
            return JSONResponse(content=json_compatible_item_data)
        except Exception as e:
            print("[Voice Changer] post_upload_file ex:", e)

    def _upload_file(self, upload_dirname: str, file: UploadFile, filename: str, index: int):
        if file and filename:
            try:
                filename = self._sanitize_filename(filename)
            except RuntimeError:
                response = EasyFileUploaderResponse(
                    status="NG",
                    message=f"filename {filename} is not allowed. Allowed filenames are {self.allowed_filenames}",
                )
                json_compatible_item_data = jsonable_encoder(response)
                return JSONResponse(content=json_compatible_item_data)

            filename = f"{filename}_{index}"
            fileobj = file.file
            target_path = os.path.join(upload_dirname, filename)
            target_dir = os.path.dirname(target_path)
            os.makedirs(target_dir, exist_ok=True)
            upload_dir = open(target_path, "wb+")
            shutil.copyfileobj(fileobj, upload_dir)
            upload_dir.close()

            response = EasyFileUploaderResponse(
                status="OK",
                message=f"uploaded files {filename} ",
            )
            json_compatible_item_data = jsonable_encoder(response)
            return JSONResponse(content=json_compatible_item_data)

        else:
            response = EasyFileUploaderResponse(
                status="NG",
                message="no file found",
            )
            json_compatible_item_data = jsonable_encoder(response)
            return JSONResponse(content=json_compatible_item_data)

    #######################################
    # Concat File
    # TEST:
    #  curl -X POST -F "filename=data1"  -F "filenameChunkNum=3" http://localhost:8001/concat_uploaded_file
    #######################################
    def post_concat_uploaded_file(self, filename: str = Form(...), filenameChunkNum: int = Form(...)):
        try:
            res = self._concat_file_chunks(self.upload_dir, filename, filenameChunkNum, self.upload_dir)
            json_compatible_item_data = jsonable_encoder(res)
            return JSONResponse(content=json_compatible_item_data)
        except Exception as e:
            print("[Voice Changer] post_concat_uploaded_file ex:", e)

    def _concat_file_chunks(self, upload_dirname: str, filename: str, chunkNum: int, dest_dirname: str):
        try:
            filename = self._sanitize_filename(filename)
        except RuntimeError:
            response = EasyFileUploaderResponse(
                status="NG",
                message=f"filename {filename} is not allowed. Allowed filenames are {self.allowed_filenames}",
            )
            json_compatible_item_data = jsonable_encoder(response)
            return JSONResponse(content=json_compatible_item_data)

        target_path = os.path.join(upload_dirname, filename)  # filenameにサブフォルダ名が入っている場合への対策として、joinしてからdirnameする <- サニタイズでサブフォルダを削除する仕様に変更したため意味はなくなった。
        target_dir = os.path.dirname(target_path)
        os.makedirs(target_dir, exist_ok=True)

        if os.path.exists(target_path):
            os.remove(target_path)
        with open(target_path, "ab") as out:
            for i in range(chunkNum):
                chunkName = f"{filename}_{i}"
                chunk_file_path = os.path.join(upload_dirname, chunkName)
                stored_chunk_file = open(chunk_file_path, "rb")
                out.write(stored_chunk_file.read())
                stored_chunk_file.close()
                os.remove(chunk_file_path)
            out.close()

        response = EasyFileUploaderResponse(
            status="OK",
            message=f"concat files {filename} ",
        )
        json_compatible_item_data = jsonable_encoder(response)
        return JSONResponse(content=json_compatible_item_data)



from fastapi import APIRouter
from typing import Callable, Dict, Any, TypeAlias, Literal

from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from colab_easy_ui.data.Response import ColabEasyUIResponse
from pydantic import BaseModel

JsonApiFuncType: TypeAlias = Literal[
    "oneshot",
    "progress",
    "get",
    "download",
]

JsonApiTaskStatus: TypeAlias = Literal[
    "RUNNING",
    "DONE",
    "NOT_FOUND",
]

JsonApiTaskProcessStatus: TypeAlias = Literal[
    "RUNNING",
    "VALIDATING",
    "DONE",
    "ERROR_INVALID_CHECKSUM",
    "ERROR_DOWNLOAD_NOT_FOUND",
]


##############
# データクラス
##############
class JsonApiFuncInfo(BaseModel):
    id: str
    type: JsonApiFuncType
    display_name: str
    method: str
    path: str


class JsonApiFunc(JsonApiFuncInfo):
    """
    JsonAPIの情報。情報を登録するときに使用する。呼びさす関数をメンバーに持つ。
    """

    func: Callable[[Dict[str, Any]], Dict[str, Any]]


class FunctionInfoResponse(ColabEasyUIResponse):
    """
    JsonAPIの情報取得APIのレスポンス
    """

    functions: list[JsonApiFuncInfo]


class FunctionTaskStatus(BaseModel):
    """
    JsonAPIのタスク情報
    """

    status: JsonApiTaskStatus
    data: Any


class GetFunctionTaskStatusResponse(ColabEasyUIResponse):
    """
    JsonAPIのタスク情報取得APIのレスポンス
    """

    task_id: str
    task_status: JsonApiTaskStatus
    task_status_data: str  # jsonの文字列。setの時にhttp経由で文字列で来るのでDict方にしていない。


class ProgressFunctionCallResponse(ColabEasyUIResponse):
    """
    progress型のJsonAPIコール時のレスポンス。uuidが返るのでこれで進捗等のステータスを確認する。
    """

    uuid: str


class GetFunctionCallResponse(ColabEasyUIResponse):
    """
    get型のJsonAPIコール時のレスポンス。
    """

    data: Dict[str, Any]


class ProgressTaskProcessStatus(BaseModel):
    """
    非同期API内の処理の進捗を管理するクラス
    """

    display_name: str
    n: int
    total: int
    status: JsonApiTaskProcessStatus
    unit: str


class ProgressTaskStatus(BaseModel):
    processStatus: Dict[str, ProgressTaskProcessStatus]


##############
# メインクラス
##############
class Functions:
    def __init__(self):
        self.status_store: Dict[str, FunctionTaskStatus] = {}

    def register_functions(self, funcs: list[JsonApiFunc]):
        self.functions = funcs

        router = APIRouter()
        for func in funcs:
            router.add_api_route(func.path, func.func, methods=[func.method])
        router.add_api_route("/functions", self.get_function_info, methods=["GET"])
        router.add_api_route("/functions_set_task_status", self.set_task_status, methods=["GET"])
        router.add_api_route("/functions_del_task_status", self.del_task_status, methods=["GET"])
        router.add_api_route("/functions_get_task_status", self.get_task_status, methods=["GET"])

        self.router = router

    def set_task_status(self, task_id: str, status: JsonApiTaskStatus, data: str):
        self.status_store[task_id] = FunctionTaskStatus(status=status, data=data)
        # print(self.status_store)

    def del_task_status(self, task_id: str):
        del self.status_store[task_id]

    def get_task_status(self, task_id: str):
        if task_id not in self.status_store:
            data = GetFunctionTaskStatusResponse(
                status="OK",
                message="",
                task_id=task_id,
                task_status="NOT_FOUND",
                task_status_data="",
            )
            json_compatible_item_data = jsonable_encoder(data)
            return JSONResponse(content=json_compatible_item_data)

        data = GetFunctionTaskStatusResponse(
            status="OK",
            message="",
            task_id=task_id,
            task_status=self.status_store[task_id].status,
            task_status_data=self.status_store[task_id].data,
        )
        json_compatible_item_data = jsonable_encoder(data)
        return JSONResponse(content=json_compatible_item_data)

    def get_function_info(self):
        try:
            if self.functions is None:
                data = ColabEasyUIResponse(
                    status="error",
                    message="No functions registered.",
                )
                return JSONResponse(content=jsonable_encoder(data))

            functions: list[JsonApiFuncInfo] = []
            for f in self.functions:
                f_info = JsonApiFuncInfo(
                    id=f.id,
                    type=f.type,
                    display_name=f.display_name,
                    method=f.method,
                    path=f.path,
                )
                functions.append(f_info)

            data = FunctionInfoResponse(
                status="OK",
                message="",
                functions=functions,
            )

            json_compatible_item_data = jsonable_encoder(data)
        except Exception as e:
            print(e)
        return JSONResponse(content=json_compatible_item_data)



from typing import Callable
from fastapi import FastAPI
from fastapi.responses import JSONResponse
from fastapi import APIRouter

# from fastapi.middleware.cors import CORSMiddleware
from fastapi import HTTPException
from fastapi import Request
from fastapi import Response
from fastapi.routing import APIRoute
from fastapi.exceptions import RequestValidationError
from fastapi.staticfiles import StaticFiles
from fastapi.encoders import jsonable_encoder

from colab_easy_ui.EasyFileUploaderInternal import EasyFileUploader
import uvicorn
import threading
import nest_asyncio
import portpicker
# from colab_easy_ui.ColabInternalFetcher import ColabInternalFetcher
from colab_easy_ui.Functions import Functions, JsonApiFunc

import logging
from colab_easy_ui.const import LOG_FILE, LOGGER_NAME
from colab_easy_ui.logger import setup_logger

setup_logger(LOGGER_NAME, LOG_FILE)


async def validation_exception_handler(request: Request, exc: RequestValidationError):
    logging.getLogger(LOGGER_NAME).info(f"Validation error occurred: {exc}")
    
    return JSONResponse(
        status_code=422,
        content=jsonable_encoder({
            "detail": exc.errors(),
            "body": (await request.body()).decode(),
        })
    )


async def log_request_middleware(request: Request, call_next):
    logging.getLogger(LOGGER_NAME).info(f"request: {request.url}")
    response = await call_next(request)
    return response


class ColabEasyUI(FastAPI):
    _instance = None
    port = 0
    server_thread = None

    @classmethod
    def get_instance(
        cls,
        port: int | None = None,
        allow_origins: list[str] = ["*"],
    ):
        if cls._instance is None:
            app_fastapi = ColabEasyUI(port=port, allow_origins=allow_origins)
            cls._instance = app_fastapi
            return cls._instance

        return cls._instance

    def __init__(self, port: int | None = None, allow_origins: list[str] = ["*"]):
        super().__init__(
            default_response_class=JSONResponse,
        )
        
        # ログ記録用ミドルウェアを追加
        self.middleware("http")(log_request_middleware)
        
        # バリデーションエラーのハンドラーを登録
        self.add_exception_handler(RequestValidationError, validation_exception_handler)
        
        if port is not None:
            self.port = port
        else:
            self.port = portpicker.pick_unused_port()

    def _run_server(self, port: int, host: str):

        config = uvicorn.Config(self, host=host, port=port, log_level="critical")
        self.server = uvicorn.Server(config)
        self.server.run()   # ブロッキング        

    def start(self, host: str = "127.0.0.1"):

        logging.getLogger(LOGGER_NAME).info(f"-----------------------------------------------")    
        logging.getLogger(LOGGER_NAME).info(f"Starting ColabEasyUI version start: {self.port}")    
        logging.getLogger(LOGGER_NAME).info(f"-----------------------------------------------")    

        nest_asyncio.apply()
        self.server_thread = threading.Thread(target=self._run_server, args=(self.port, host))
        self.server_thread.start()
        return self.port

    def stop(self):
        if self.server and self.server.should_exit is False:
            self.server.should_exit = True  # サーバーに終了要求を出す


        if self.server_thread:
            self.server_thread.join()
            self.server_thread = None

    def mount_static_folder(self, path: str, real_path: str):
        self.mount(
            path,
            StaticFiles(directory=real_path, html=True),
            name="static",
        )

    def enable_file_uploader(self, upload_dir: str, allowed_files: dict[str, str] | None = None):
        self.fileUploader = EasyFileUploader(upload_dir)
        self.fileUploader.set_allowed_filenames(allowed_files)
        self.include_router(self.fileUploader.router)

    # def enable_colab_internal_fetcher(
    #     self,
    #     project_dir: str,
    #     ipython=None,
    #     logfile=None,
    # ):
    #     from fastapi import APIRouter

    #     self.colabInternalFetcher = ColabInternalFetcher("trainer", ipython, logfile)

    #     router = APIRouter()
    #     router.add_api_route("/internal_start_tb", self.colabInternalFetcher.start_tensorboard, methods=["GET"])
    #     router.add_api_route("/internal_runs", self.colabInternalFetcher.get_runs, methods=["GET"])
    #     router.add_api_route("/internal_scalars_tags", self.colabInternalFetcher.get_scalars_tags, methods=["GET"])
    #     router.add_api_route("/internal_scalars_scalars", self.colabInternalFetcher.get_scalars_scalars, methods=["GET"])
    #     router.add_api_route("/internal_scalars_multirun", self.colabInternalFetcher.get_scalars_scalars, methods=["GET"])
    #     router.add_api_route("/internal_images_tags", self.colabInternalFetcher.get_images_tags, methods=["GET"])
    #     router.add_api_route("/internal_images_images", self.colabInternalFetcher.get_images_images, methods=["GET"])
    #     router.add_api_route("/internal_images_individualImage", self.colabInternalFetcher.get_images_individualImage, methods=["GET"])
    #     router.add_api_route("/internal_audio_tags", self.colabInternalFetcher.get_audio_tags, methods=["GET"])
    #     router.add_api_route("/internal_audio_audio", self.colabInternalFetcher.get_audio_audio, methods=["GET"])
    #     router.add_api_route("/internal_audio_individualAudio", self.colabInternalFetcher.get_audio_individualAudio, methods=["GET"])
    #     self.include_router(router)

    def register_functions(self, funcs: list[JsonApiFunc]):
        functions = Functions()
        functions.register_functions(funcs)
        self.include_router(functions.router)

